import argparse
import asyncio
from datetime import datetime
import json
import logging
import sys
import threading
from typing import List

from . import Charger, Circuit, DatatypesStreamData, Easee, Equalizer, Site
from .utils import lookup_charger_stream_id, lookup_equalizer_stream_id

CACHED_TOKEN = "easee-token.json"

_LOGGER = logging.getLogger(__file__)


def add_input(queue):
    queue.put_nowait(sys.stdin.read(1))


async def print_signalr(id, data_type, data_id, value):

    type_str = DatatypesStreamData(data_type).name
    if id[0] == "Q":
        data_str = lookup_equalizer_stream_id(data_id)
    else:
        data_str = lookup_charger_stream_id(data_id)

    print(f"SR: {id} data type {data_type} {type_str} data id {data_id} {data_str} value {value}")


def parse_arguments():
    parser = argparse.ArgumentParser(description="Read data from your Easee EV installation")
    parser.add_argument("-u", "--username", help="Username", required=True)
    parser.add_argument("-p", "--password", help="Password", required=True)
    parser.add_argument("-c", "--chargers", help="Get chargers information", action="store_true")
    parser.add_argument("-s", "--sites", help="Get sites information", action="store_true")
    parser.add_argument("-ci", "--circuits", help="Get circuits information", action="store_true")
    parser.add_argument("-e", "--equalizers", help="Get equalizers information", action="store_true")
    parser.add_argument(
        "-a",
        "--all",
        help="Get all sites, circuits, equalizers and chargers information",
        action="store_true",
    )
    parser.add_argument(
        "-sum",
        "--summary",
        help="Get summary of sites, circuits, equalizers and chargers information",
        action="store_true",
    )
    parser.add_argument(
        "-f", "--force", help="Force update of lifetime energy and charger opmode ", action="store_true"
    )
    parser.add_argument("-l", "--loop", help="Loop charger data every 5 seconds", action="store_true")
    parser.add_argument("-r", "--signalr", help="Listen to signalr stream", action="store_true")
    parser.add_argument("-co", "--cost", help="Retrieve cost for last year", action="store_true")
    parser.add_argument("--countries", help="Get active countries information", action="store_true")
    parser.add_argument(
        "-d",
        "--debug",
        help="Print debugging statements",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.WARNING,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Be verbose",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )
    args = parser.parse_args()
    logging.basicConfig(
        format="%(asctime)-15s %(name)-5s %(levelname)-8s %(message)s",
        level=args.loglevel,
    )
    return args


# TODO: Add option to send in a cached token
# def token_read():
#     try:
#         with open(CACHED_TOKEN, "r") as token_file:
#             return json.load(token_file)
#     except FileNotFoundError:
#         return None


# def token_write(token):
#     with open(CACHED_TOKEN, "w") as token_file:
#         json.dump(token, token_file, indent=2)


async def async_main():
    args = parse_arguments()
    _LOGGER.debug("args: %s", args)
    easee = Easee(args.username, args.password)

    if args.chargers:
        chargers: List[Charger] = await easee.get_chargers()
        await chargers_info(chargers)

    if args.sites:
        sites: List[Site] = await easee.get_account_products()
        await sites_info(sites)

    if args.circuits:
        sites: List[Site] = await easee.get_sites()
        for site in sites:
            await circuits_info(circuits=site.get_circuits())

    if args.equalizers:
        sites: List[Site] = await easee.get_sites()
        for site in sites:
            await equalizers_info(equalizers=site.get_equalizers())

    if args.countries:
        countries_active = await easee.get_active_countries()
        print(json.dumps(countries_active, indent=2))

    if args.cost:
        dt_end = datetime.now()
        dt_start = datetime.now().replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        sites: List[Site] = await easee.get_sites()
        for site in sites:
            costs = await site.get_cost_between_dates(dt_start, dt_end)
            await costs_info(costs)

    if args.all:
        sites: List[Site] = await easee.get_sites()
        await sites_info(sites)
        for site in sites:
            equalizers = site.get_equalizers()
            await equalizers_info(equalizers)
            circuits = site.get_circuits()
            await circuits_info(circuits)
            for circuit in circuits:
                chargers = circuit.get_chargers()
                await chargers_info(chargers)

    if args.summary:
        sites: List[Site] = await easee.get_sites()
        circuits = List[Circuit]
        chargers = List[Charger]
        equalizers = List[Equalizer]

        for site in sites:
            print(
                f"  "
                f" Site: {site.__getitem__('name')}"
                f" (ID: {site.id}),"
                f" {site.__getitem__('address')['street']},"
                f" main fuse {site.__getitem__('ratedCurrent')}A"
                f" "
            )
            equalizers = site.get_equalizers()
            for equalizer in equalizers:
                print(
                    f"    "
                    f" Equalizer: #{equalizer.__getitem__('name')}"
                    f" (ID: {equalizer.id})"
                    f" SiteID: #{equalizer.__getitem__('siteId')}"
                    f" CircuitId: #{equalizer.__getitem__('circuitId')}"
                    f" "
                )
            circuits = site.get_circuits()
            for circuit in circuits:
                print(
                    f"    "
                    f" Circuit: #{circuit.__getitem__('circuitPanelId')}"
                    f" {circuit.__getitem__('panelName')}"
                    f" (ID: {circuit.id})"
                    f" {circuit.__getitem__('ratedCurrent')}A"
                    f" "
                )
                chargers = circuit.get_chargers()
                for charger in chargers:
                    state = await charger.get_state()
                    config = await charger.get_config()
                    print(
                        f"      "
                        f" Charger: {charger.__getitem__('name')}"
                        f" (ID: {charger.id}),"
                        f" enabled: {config.__getitem__('isEnabled')}"
                        f" online: {state.__getitem__('isOnline')}"
                        f" version: {state.__getitem__('chargerFirmware')}"
                        f" voltage: {round(state.__getitem__('voltage'),1)}"
                        f" current: {round(state.__getitem__('outputCurrent'),1)}"
                        f" "
                    )

        print(f"\n\nFound {len(sites)} site(s), {len(circuits)} circuit(s) and {len(chargers)} charger(s).")

    if args.loop:
        sites: List[Site] = await easee.get_sites()
        for site in sites:
            circuits = site.get_circuits()
            for circuit in circuits:
                chargers = circuit.get_chargers()
                try:
                    header = True
                    while True:
                        for charger in chargers:
                            await charger_loop(charger, header)
                            header = False
                            await asyncio.sleep(5)
                except KeyboardInterrupt as e:  # noqa
                    # Close connection on user interuption
                    print("Interrupted by user")
                    await easee.close()
                except Exception as e:
                    print(e)
                    await easee.close()

    if args.force:
        chargers: List[Charger] = await easee.get_chargers()
        for charger in chargers:
            print(f"Forcing update of lifetimeenergy on charger {charger['id']}")
            await charger.force_update_lifetimeenergy()
            print(f"Forcing update of opmode on charger {charger['id']}")
            await charger.force_update_opmode()

    if args.signalr:
        chargers: List[Charger] = await easee.get_chargers()
        equalizers = []
        sites: List[Site] = await easee.get_sites()
        for site in sites:
            equalizers_site = site.get_equalizers()
            for equalizer in equalizers_site:
                equalizers.append(equalizer)
        for charger in chargers:
            await easee.sr_subscribe(charger, print_signalr)
        for equalizer in equalizers:
            await easee.sr_subscribe(equalizer, print_signalr)

        queue = asyncio.Queue(1)
        input_thread = threading.Thread(target=add_input, args=(queue,))
        input_thread.daemon = True
        input_thread.start()

        while True:
            await asyncio.sleep(1)

            if queue.empty() is False:
                #                print "\ninput:", input_queue.get()
                break

    await easee.close()


async def chargers_info(chargers: List[Charger]):
    print("\n\n****************\nCHARGERS\n****************")
    data = []
    for charger in chargers:
        state = await charger.get_state()
        config = await charger.get_config()
        schedule = await charger.get_basic_charge_plan()
        week_schedule = await charger.get_weekly_charge_plan()
        observation_test = await charger.get_observations(30, 31, 35, 36, 45)
        firmware = await charger.get_latest_firmware()
        ch = charger.get_data()
        ch["state"] = state.get_data()
        ch["config"] = config.get_data()
        ch["firmware"] = firmware
        ch["observation"] = observation_test
        if schedule is not None:
            ch["schedule"] = schedule.get_data()
        if week_schedule is not None:
            ch["week_schedule"] = week_schedule.get_data()
        data.append(ch)

    print(json.dumps(data, indent=2))


async def sites_info(sites: List[Site]):
    print("\n\n****************\nSITES\n****************")
    data = []
    for site in sites:
        data.append(site.get_data())

    print(json.dumps(data, indent=2))


async def circuits_info(circuits: List[Circuit]):
    print("\n\n****************\nCIRCUITS\n****************")
    data = []
    for circuit in circuits:
        data.append(circuit.get_data())

    print(json.dumps(data, indent=2))


async def equalizers_info(equalizers: List[Equalizer]):
    print("\n\n****************\nEQUALIZERS\n****************")
    data = []
    for equalizer in equalizers:
        eq = equalizer.get_data()
        state = await equalizer.get_state()
        config = await equalizer.get_config()
        eq["state"] = state.get_data()
        eq["config"] = config.get_data()
        data.append(eq)

    print(
        json.dumps(
            data,
            indent=2,
        )
    )


async def costs_info(costs):
    print("\n\n****************\nCOST\n****************")
    data = []
    for cost in costs:
        data.append(cost["chargerId"])
        data.append(cost["totalCost"])
        data.append(cost["currencyId"])
        data.append(cost["totalEnergyUsage"])

    print(json.dumps(data, indent=2))


async def charger_loop(charger: Charger, header=False):
    """Return the state attributes."""
    # await charger.async_update()
    state = await charger.get_state()
    # config = await charger.get_config() # not used yet

    if header:
        print(str_fixed_length("NAME", 15), end=" ")
        print(str_fixed_length("OPMODE", 20), end=" ")
        print(str_fixed_length("ONLINE", 7), end=" ")
        print(str_fixed_length("POWER", 7), end=" ")
        print(str_fixed_length("OUTCURR", 10), end=" ")
        print(str_fixed_length("IN_T2", 10), end=" ")
        print(str_fixed_length("IN_T3", 10), end=" ")
        print(str_fixed_length("IN_T4", 10), end=" ")
        print(str_fixed_length("IN_T5", 10), end=" ")
        print(str_fixed_length("VOLTAGE", 10), end=" ")
        print(str_fixed_length("kWh", 10), end=" ")
        print(str_fixed_length("RATE", 10), end=" ")
        print(str_fixed_length("REASON", 25), end=" ")
        print(" ")

    print(str_fixed_length(f"{charger.name}", 15), end=" ")
    print(str_fixed_length(f"{state.__getitem__('chargerOpMode')}", 20), end=" ")
    print(str_fixed_length(f"{state.__getitem__('isOnline')}", 7), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('totalPower'),2)}kW", 7), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('outputCurrent'),1)}A", 10), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('inCurrentT2'),1)}A", 10), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('inCurrentT3'),1)}A", 10), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('inCurrentT4'),1)}A", 10), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('inCurrentT5'),1)}A", 10), end=" ")
    print(str_fixed_length(f"{round(state.__getitem__('voltage'),1)}V", 10), end=" ")
    print(
        str_fixed_length(f"{round(state.__getitem__('sessionEnergy'),2)}kWh", 10),
        end=" ",
    )
    print(
        str_fixed_length(f"{round(state.__getitem__('energyPerHour'),2)}kWh/h", 10),
        end=" ",
    )
    print(str_fixed_length(f"{str(state.__getitem__('reasonForNoCurrent'))}", 25), end=" ")
    print(" ")


def str_fixed_length(myStr, length: int):
    while len(myStr) < length:
        myStr = myStr + " "
    if len(myStr) > length:
        myStr = myStr[0:length]
    return myStr


def main():
    asyncio.run(async_main())


if __name__ == "__main__":
    import time

    s = time.perf_counter()
    main()
    elapsed = time.perf_counter() - s
    print(f"{__file__} executed in {elapsed:0.2f} seconds.")



"""
Main client for the Eesee account.
"""
import asyncio
from datetime import datetime, timedelta
import logging
from typing import Any, Dict, List

import aiohttp
from pysignalr.client import SignalRClient
from pysignalr.exceptions import AuthorizationError
from pysignalr.messages import CompletionMessage

from .charger import Charger
from .exceptions import (
    AuthorizationFailedException,
    BadRequestException,
    ForbiddenServiceException,
    NotFoundException,
    ServerFailureException,
    TooManyRequestsException,
)
from .site import Site, SiteState
from .utils import convert_stream_data

__VERSION__ = "0.8.1"

_LOGGER = logging.getLogger(__name__)

SR_MIN_BACKOFF = 0
SR_MAX_BACKOFF = 300
SR_INC_BACKOFF = 30


async def raise_for_status(response):
    if 400 <= response.status:
        e = aiohttp.ClientResponseError(
            response.request_info,
            response.history,
            code=response.status,
            headers=response.headers,
        )

        if "json" in response.headers.get("CONTENT-TYPE", ""):
            data = await response.json()
            e.message = str(data)
        else:
            data = await response.text()

        if 400 == response.status:
            _LOGGER.debug("Bad request " + f"({response.status}: {data} {response.url})")
            raise BadRequestException(data)
        elif 401 == response.status:
            _LOGGER.debug("Unautorized " + f"({response.status}: {data} {response.url})")
            raise AuthorizationFailedException(data)
        elif 403 == response.status:
            _LOGGER.debug("Forbidden service" + f"({response.status}: {response.url})")
            raise ForbiddenServiceException(data)
        elif 404 == response.status:
            # Getting this error when getting or deleting charge schedules which doesn't exist (empty)
            _LOGGER.debug("Not found " + f"({response.status}: {data} {response.url})")
            raise NotFoundException(data)
        elif 429 == response.status:
            _LOGGER.debug("Too many requests " + f"({response.status}: {data} {response.url})")
            raise TooManyRequestsException(data)
        elif response.status >= 500:
            _LOGGER.warning("Server failure" + f"({response.status}: {response.url})")
            raise ServerFailureException(data)
        else:
            _LOGGER.error("Error in request to Easee API: %s", data)
            raise Exception(data) from e
        return False
    else:
        return True


class Easee:
    def __init__(self, username, password, session: aiohttp.ClientSession = None):
        self.username = username
        self.password = password
        self.external_session = True if session else False

        _LOGGER.info("Easee python library version: %s", __VERSION__)

        self.base = "https://api.easee.com"
        self.sr_base = "https://streams.easee.com/hubs/chargers"
        self.token = {}
        self.headers = {
            "User-Agent": f"pyeasee/{__VERSION__} REST client",
            "Accept": "application/json",
            "Content-Type": "application/json;charset=UTF-8",
        }
        self.minimal_headers = self.headers
        self.sr_headers = {
            "User-Agent": f"pyeasee/{__VERSION__} SignalR client",
            "Accept": "application/json",
            "Content-Type": "application/json;charset=UTF-8",
        }
        if session is None:
            self.session = aiohttp.ClientSession()
        else:
            self.session = session

        self.sr_subscriptions = {}
        self.sr_connection = None
        self.sr_connected = False
        self.sr_connect_in_progress = False
        self._sr_backoff = SR_MIN_BACKOFF
        self._sr_task = None

    def base_uri(self):
        return self.base

    async def post(self, url, **kwargs):
        _LOGGER.debug("POST: %s (%s)", url, kwargs)
        await self._verify_updated_token()
        response = await self.session.post(f"{self.base}{url}", headers=self.headers, **kwargs)
        await self.check_status(response)
        return response

    async def put(self, url, **kwargs):
        _LOGGER.debug("PUT: %s (%s)", url, kwargs)
        await self._verify_updated_token()
        response = await self.session.put(f"{self.base}{url}", headers=self.headers, **kwargs)
        await self.check_status(response)
        return response

    async def get(self, url, **kwargs):
        _LOGGER.debug("GET: %s (%s)", url, kwargs)
        await self._verify_updated_token()
        response = await self.session.get(f"{self.base}{url}", headers=self.headers, **kwargs)
        await self.check_status(response)
        return response

    async def delete(self, url, **kwargs):
        _LOGGER.debug("DELETE: %s (%s)", url, kwargs)
        await self._verify_updated_token()
        response = await self.session.delete(f"{self.base}{url}", headers=self.headers, **kwargs)
        await self.check_status(response)
        return response

    # TODO: Quick fix for the auth fail errors due to something wrong with refresh token logic
    async def check_status(self, response):
        try:
            await raise_for_status(response)
        except (AuthorizationFailedException, BadRequestException):
            _LOGGER.debug("Re authorizing due to 401")
            await self.connect()
            # rethrow it
            await raise_for_status(response)
        except ForbiddenServiceException:
            raise
        except Exception as ex:
            _LOGGER.debug("Got other exception from status: %s", type(ex).__name__)
            raise

    async def _verify_updated_token(self):
        """
        Make sure there is a valid token
        """
        if "accessToken" not in self.token:
            await self.connect()
        _LOGGER.debug(
            "verify_updated_token: %s, %s, %s",
            self.token["expires"],
            datetime.now(),
            self.token["expires"] < datetime.now(),
        )
        if self.token["expires"] < datetime.now():
            await self._refresh_token()
        accessToken = self.token["accessToken"]
        self.headers["Authorization"] = f"Bearer {accessToken}"
        self.sr_headers["Authorization"] = f"Bearer {accessToken}"

    async def _handle_token_response(self, res):
        """
        Handle the token request and set new datetime when it expires
        """
        self.token = await res.json()
        _LOGGER.debug("TOKEN: %s", self.token)
        expiresIn = int(self.token["expiresIn"]) - 60
        now = datetime.now()
        self.token["expires"] = now + timedelta(0, expiresIn)

    async def connect(self):
        """
        Connect and gets initial token
        """
        data = {"userName": self.username, "password": self.password}
        _LOGGER.debug("getting token for user: %s", self.username)
        response = await self.session.post(f"{self.base}/api/accounts/login", headers=self.minimal_headers, json=data)
        await raise_for_status(response)
        await self._handle_token_response(response)

    async def _refresh_token(self):
        """
        Refresh token
        """
        data = {
            "accessToken": self.token["accessToken"],
            "refreshToken": self.token["refreshToken"],
        }
        _LOGGER.debug("Refreshing access token")
        try:
            res = await self.session.post(
                f"{self.base}/api/accounts/refresh_token", headers=self.minimal_headers, json=data
            )
            await self._handle_token_response(res)
        except (AuthorizationFailedException, BadRequestException):
            _LOGGER.debug("Could not get new access token from refresh token, getting new one")
            await self.connect()

    async def close(self):
        """
        Close the underlying aiohttp session
        """
        if self.session and self.external_session is False:
            await self.session.close()
            self.session = None

        await self._sr_disconnect()

    def _sr_next(self):
        if self._sr_backoff < SR_MAX_BACKOFF:
            self._sr_backoff = self._sr_backoff + SR_INC_BACKOFF
        return self._sr_backoff

    async def _sr_open_cb(self) -> None:
        """
        Signalr connected callback - called from signalr thread, internal use only
        """
        _LOGGER.debug("SignalR stream connected")
        self._sr_backoff = SR_MIN_BACKOFF
        self.sr_connected = True

        for id in self.sr_subscriptions:
            _LOGGER.debug("Subscribing to %s", id)
            await self.sr_connection.send("SubscribeWithCurrentState", [id, True])

    async def _sr_close_cb(self) -> None:
        """
        Signalr disconnected callback - called from signalr thread, internal use only
        """
        _LOGGER.debug("SignalR stream disconnected")
        self.sr_connection = None
        self.sr_connected = False
        await self._sr_connect(SR_INC_BACKOFF)

    async def _sr_error_cb(self, message: CompletionMessage) -> None:
        _LOGGER.debug("SignalR error recevied {message.error}")

    async def _sr_product_update_cb(self, stuff: List[Dict[str, Any]]) -> None:
        """
        Signalr new data recieved callback - called from signalr thread, internal use only
        """
        for data in stuff:
            await self._sr_callback(data)

    async def _sr_callback(self, stuff: List[Dict[str, Any]]):
        """
        Signalr callback handler - internal use only
        """
        if stuff["mid"] in self.sr_subscriptions:
            callback = self.sr_subscriptions[stuff["mid"]]
            value = convert_stream_data(stuff["dataType"], stuff["value"])
            await callback(stuff["mid"], stuff["dataType"], stuff["id"], value)
        else:
            _LOGGER.debug("No callback found for %s", stuff["mid"])

    async def _sr_connect(self, start_delay=0):
        """
        Signalr connect - internal use only
        """
        if self.sr_connect_in_progress is True:
            _LOGGER.debug("Already connecting")
            return
        self.sr_connect_in_progress = True

        _LOGGER.debug(f"SR connect sleep {start_delay}")
        await asyncio.sleep(start_delay)

        self._sr_task = asyncio.create_task(self._sr_connect_loop())

    async def _sr_connect_loop(self):
        """
        Signalr connect loop - internal use only
        """
        if self.sr_connection is not None:
            return

        _LOGGER.debug("SR connect loop")

        while True:
            try:
                await self._verify_updated_token()
                self.sr_connection = SignalRClient(self.sr_base, headers=self.sr_headers)
                self.sr_connection.on_open(self._sr_open_cb)
                self.sr_connection.on_close(self._sr_close_cb)
                self.sr_connection.on_error(self._sr_error_cb)
                self.sr_connection.on("ProductUpdate", self._sr_product_update_cb)
                await self.sr_connection.run()
            except AuthorizationError as ex:
                backoff = self._sr_next()
                _LOGGER.debug("SR authentication failed: %s. Retry in %d seconds", type(ex).__name__, backoff)
                await asyncio.sleep(backoff)
                await self._refresh_token()
                continue
            except Exception as ex:
                backoff = self._sr_next()
                _LOGGER.debug("SR start exception: %s: %s. Retry in %d seconds", type(ex).__name__, ex, backoff)
                await asyncio.sleep(backoff)
                continue

            break

        self.sr_connect_in_progress = False

    def sr_is_connected(self):
        return self.sr_connected

    async def sr_subscribe(self, product, callback):
        """
        Subscribe to signalr events for product, callback will be called as async callback(product_id, data_type, data_id, value)
        """
        if product.id in self.sr_subscriptions:
            return

        _LOGGER.debug("Subscribing to %s", product.id)
        self.sr_subscriptions[product.id] = callback
        if self.sr_connected is True:
            if self.running_loop is not None:
                await self.sr_connection.send("SubscribeWithCurrentState", [product.id, True])
        else:
            await self._sr_connect()

    async def sr_unsubscribe(self, product):
        """
        Unsubscribe from signalr events for product
        BUG: Does not work
        """
        _LOGGER.debug("Unsubscribing from %s", product.id)
        if product.id in self.sr_subscriptions:
            del self.sr_subscriptions[product.id]
            await self._sr_disconnect()
            await self._sr_connect()

    async def _sr_disconnect(self):
        """
        Disconnect the signalr stream - internal use only
        """
        if self.sr_connection is not None:
            self.sr_connection = None

    async def get_chargers(self) -> List[Charger]:
        """
        Retrieve all chargers
        """
        try:
            records = await (await self.get("/api/chargers")).json()
            _LOGGER.debug("Chargers:  %s", records)
            return [Charger(k, self) for k in records]
        except (ServerFailureException):
            return None

    async def get_site(self, id: int) -> Site:
        """get site by id"""
        try:
            data = await (await self.get(f"/api/sites/{id}?detailed=true")).json()
            _LOGGER.debug("Site:  %s", data)
            return Site(data, self)
        except (ServerFailureException):
            return None

    async def get_sites(self) -> List[Site]:
        """Get all sites"""
        try:
            records = await (await self.get("/api/sites")).json()
            _LOGGER.debug("Sites:  %s", records)
            sites = await asyncio.gather(*[self.get_site(r["id"]) for r in records])
            return sites
        except (ServerFailureException):
            return None

    async def get_account_products(self) -> List[Site]:
        """Get all sites and products that are accessible by the logged in user"""
        try:
            records = await (await self.get("/api/accounts/products")).json()
            _LOGGER.debug("Sites:  %s", records)
            sites = []
            for r in records:
                site = await self.get_site(r["id"])
                site["circuits"] = r["circuits"]
                site["equalizers"] = r["equalizers"]
                sites.append(site)
            return sites
        except (ServerFailureException):
            return None

    async def get_site_state(self, id: str) -> SiteState:
        """Get site state"""
        try:
            state = await (await self.get(f"/api/sites/{id}/state")).json()
            return SiteState(state)
        except (ServerFailureException):
            return None

    async def get_active_countries(self) -> List[Any]:
        """Get all active countries"""
        try:
            records = await (await self.get("/api/resources/countries/active")).json()
            _LOGGER.debug("Active countries:  %s", records)
            return records
        except (ServerFailureException):
            return None

    async def get_currencies(self) -> List[Any]:
        """Get all currencies"""
        try:
            records = await (await self.get("/api/resources/currencies")).json()
            _LOGGER.debug("Currencies:  %s", records)
            return records
        except (ServerFailureException):
            return None



class AuthorizationFailedException(Exception):
    pass


class NotFoundException(Exception):
    pass


class TooManyRequestsException(Exception):
    pass


class ServerFailureException(Exception):
    pass


class ExceptionWithDict(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return self.message.get("title", "")
        else:
            return f"{type(self).__name__} has been raised"


class ForbiddenServiceException(ExceptionWithDict):
    """Authenticated but access to resource not allowed."""


class BadRequestException(ExceptionWithDict):
    """Bad arguments or operation not allowed in this context."""



# Data IDs for Charger from signalr stream

from enum import Enum


class ChargerStreamData(Enum):
    state_selfTestResult = 1  # OK or error codes [String] ['Admin', 'Partner', 'User']
    SelfTestDetails = 2  # JSON with details from self-test [String] ['Admin', 'Partner']
    WifiEvent = 10  # Enum with WiFi event codes. Requires telemetry debug mode. Will be updated on WiFi events when using cellular, will otherwise be reported in ChargerOfflineReason [Integer] ['Admin', 'Partner']
    state_chargerOfflineReason = 11  # Enum describing why charger is offline [Integer] ['Admin', 'Partner', 'User']
    EaseeLinkCommandResponse = (
        13  # Response on a EaseeLink command sent to another devic [Integer] ['Admin', 'Partner']
    )
    EaseeLinkDataReceived = 14  # Data received on EaseeLink from another device [String] ['Admin', 'Partner']
    config_localPreAuthorizeEnabled = (
        15  # Preauthorize with whitelist enabled. Readback on setting [event] [Boolean] ['Admin', 'Partner', 'User']
    )
    config_localAuthorizeOfflineEnabled = 16  # Allow offline charging for whitelisted RFID token. Readback on setting [event] [Boolean] ['Admin', 'Partner', 'User']
    config_allowOfflineTxForUnknownId = 17  # Allow offline charging for all RFID tokens. Readback on setting [event] [Boolean] ['Admin', 'Partner', 'User']
    ErraticEVMaxToggles = 18  # 0 == erratic checking disabled, otherwise the number of toggles between states Charging and Charging Complate that will trigger an error [Integer] ['Admin', 'Partner']
    BackplateType = 19  # Readback on backplate type [Integer] ['Admin', 'Partner']
    SiteStructure = 20  # Site Structure [boot] [String] ['Admin', 'Partner']
    config_detectedPowerGridType = (
        21  # Detected power grid type according to PowerGridType table [boot] [Integer] ['Admin', 'Partner', 'User']
    )
    config_circuitMaxCurrentP1 = 22  # Set circuit maximum current [Amperes] [Double] ['Admin', 'Partner', 'User']
    config_circuitMaxCurrentP2 = 23  # Set circuit maximum current [Amperes] [Double] ['Admin', 'Partner', 'User']
    config_circuitMaxCurrentP3 = 24  # Set circuit maximum current [Amperes] [Double] ['Admin', 'Partner', 'User']
    Location = 25  # Location coordinate [event] [Position] ['Admin', 'Partner']
    SiteIDString = 26  # Site ID string [event] [String] ['Admin', 'Partner']
    SiteIDNumeric = 27  # Site ID numeric value [event] [Integer] ['Admin', 'Partner']
    RfidAuthTimeoutSec = 28  # Timeout for backend to send authorization reply. After timeout offline rules apply. [Integer] ['Admin', 'Partner']
    state_lockCablePermanently = 30  # Lock type2 cable permanently [Boolean] ['Admin', 'Partner', 'User']
    config_isEnabled = 31  # Set true to enable charger, false disables charger [Boolean] ['Admin', 'Partner', 'User']
    state_temperatureMonitorState = (
        32  # State of the monitor: OFF=0, MONITORING=-1, ACTIVE=1 [Integer] ['Admin', 'Partner']
    )
    CircuitSequenceNumber = 33  # Charger sequence number on circuit [Integer] ['Admin', 'Partner']
    SinglePhaseNumber = 34  # Phase to use in 1-phase charging [Integer] ['Admin', 'Partner']
    config_enable3PhasesDEPRECATED = 35  # Allow charging using 3-phases [Boolean] ['Admin', 'Partner', 'User']
    config_wiFiSSID = 36  # WiFi SSID name [String] ['Admin', 'Partner', 'User']
    config_enableIdleCurrent = 37  # Charger signals available current when EV is done charging [user option][event] [Boolean] ['Admin', 'Partner', 'User']
    config_phaseMode = 38  # Phase mode on this charger. 1-Locked to 1-Phase, 2-Auto, 3-Locked to 3-phase(only Home) [Integer] ['Admin', 'Partner', 'User']
    ForcedThreePhaseOnITWithGndFault = 39  # Default disabled. Must be set manually if grid type is indeed three phase IT [Boolean] ['Admin', 'Partner']
    config_ledStripBrightness = 40  # LED strip brightness, 0-100% [Integer] ['Admin', 'Partner', 'User']
    config_localAuthorizationRequired = 41  # Local RFID authorization is required for charging [user options][event] [Boolean] ['Admin', 'Partner', 'User']
    config_authorizationRequired = 42  # Authorization is requried for charging [Boolean] ['Admin', 'Partner', 'User']
    config_remoteStartRequired = 43  # Remote start required flag [event] [Boolean] ['Admin', 'Partner', 'User']
    config_smartButtonEnabled = 44  # Smart button is enabled [Boolean] ['Admin', 'Partner', 'User']
    config_offlineChargingMode = 45  # Charger behavior when offline [Integer] ['Admin', 'Partner', 'User']
    state_ledMode = 46  # Charger LED mode [event] [Integer] ['Admin', 'Partner', 'User']
    config_maxChargerCurrent = 47  # Max current this charger is allowed to offer to car (A). Non volatile. [Double] ['Admin', 'Partner', 'User']
    state_dynamicChargerCurrent = (
        48  # Max current this charger is allowed to offer to car (A). Volatile [Double] ['Admin', 'Partner', 'User']
    )
    state_offlineMaxCircuitCurrentP1 = (
        50  # Maximum circuit current P1 when offline [event] [Integer] ['Admin', 'Partner', 'User']
    )
    state_offlineMaxCircuitCurrentP2 = (
        51  # Maximum circuit current P2 when offline [event] [Integer] ['Admin', 'Partner', 'User']
    )
    state_offlineMaxCircuitCurrentP3 = (
        52  # Maximum circuit current P3 when offline [event] [Integer] ['Admin', 'Partner', 'User']
    )
    ReleaseCableAtPowerOff = 54  # Release Cable At Power Off [Boolean] ['Admin', 'Partner']
    ListenToControlPulse = 56  # True = charger needs control pulse to consider itself online. Readback on charger setting [event] [Boolean] ['Admin', 'Partner']
    ControlPulseRTT = 57  # Control pulse round-trip time in milliseconds [Integer] ['Admin', 'Partner']
    schedule_chargingSchedule = 62  # Charging schedule [json] [String] ['Admin', 'Partner', 'User']
    config_pairedEqualizer = 65  # Paired equalizer details [String] ['Admin', 'Partner']
    state_wiFiAPEnabled = (
        68  # True if WiFi Access Point is enabled, otherwise false [Boolean] ['Admin', 'Partner', 'User']
    )
    state_pairedUserIDToken = (
        69  # Observed user token when charger put in RFID pairing mode [event] [String] ['Admin', 'Partner', 'User']
    )
    state_circuitTotalAllocatedPhaseConductorCurrentL1 = 70  # Total current allocated to L1 by all chargers on the circuit. Sent in by master only [Double] ['Admin', 'Partner', 'User']
    state_circuitTotalAllocatedPhaseConductorCurrentL2 = 71  # Total current allocated to L2 by all chargers on the circuit. Sent in by master only [Double] ['Admin', 'Partner', 'User']
    state_circuitTotalAllocatedPhaseConductorCurrentL3 = 72  # Total current allocated to L3 by all chargers on the circuit. Sent in by master only [Double] ['Admin', 'Partner', 'User']
    state_circuitTotalPhaseConductorCurrentL1 = 73  # Total current in L1 (sum of all chargers on the circuit) Sent in by master only [Double] ['Admin', 'Partner', 'User']
    state_circuitTotalPhaseConductorCurrentL2 = 74  # Total current in L2 (sum of all chargers on the circuit) Sent in by master only [Double] ['Admin', 'Partner', 'User']
    state_circuitTotalPhaseConductorCurrentL3 = 75  # Total current in L3 (sum of all chargers on the circuit) Sent in by master only [Double] ['Admin', 'Partner', 'User']
    state_numberOfCarsConnected = 76  # Number of cars connected to this circuit [Integer] ['Admin', 'Partner']
    state_numberOfCarsCharging = 77  # Number of cars currently charging [Integer] ['Admin', 'Partner']
    state_numberOfCarsInQueue = (
        78  # Number of cars currently in queue, waiting to be allocated power [Integer] ['Admin', 'Partner']
    )
    state_numberOfCarsFullyCharged = 79  # Number of cars that appear to be fully charged [Integer] ['Admin', 'Partner']
    state_chargerFirmware = 80  # Embedded software package release id [boot] [Integer] ['Admin', 'Partner', 'User']
    state_ICCID = 81  # SIM integrated circuit card identifier [String] ['Admin', 'Partner']
    ModemFwId = 82  # Modem firmware version [String] ['Admin', 'Partner']
    OTAErrorCode = 83  # OTA error code, see table [event] [Integer] ['Admin', 'Partner']
    MobileNetworkOperator = 84  # Current mobile network operator [pollable] [String] ['Admin', 'Partner']
    RebootReason = 89  # Reason of reboot. Bitmask of flags. [Integer] ['Admin', 'Partner']
    PowerPCBVersion = 90  # Power PCB hardware version [Integer] ['Admin', 'Partner']
    ComPCBVersion = 91  # Communication PCB hardware version [Integer] ['Admin', 'Partner']
    state_reasonForNoCurrent = 96  # Enum describing why a charger with a car connected is not offering current to the car [Integer] ['Admin', 'Partner', 'User']
    LoadBalancingNumberOfConnectedChargers = 97  # Number of connected chargers in the load balancin. Including the master. Sent from Master only. [Integer] ['Admin', 'Partner']
    state_UDPNumOfConnectedNodes = (
        98  # Number of chargers connected to master through UDP / WIFI [Integer] ['Admin', 'Partner']
    )
    state_localConnection = 99  # Slaves only. Current connection to master, 0 = None, 1= Radio, 2 = WIFI UDP, 3 = Radio and WIFI UDP [Integer] ['Admin', 'Partner']
    state_pilotMode = 100  # Pilot Mode Letter (A-F) [event] [String] ['Admin', 'Partner']
    CarConnectedDEPRECATED = 101  # Car connection state [Boolean] ['Admin', 'Partner']
    state_smartCharging = (
        102  # Smart charging state enabled by capacitive touch button [event] [Boolean] ['Admin', 'Partner', 'User']
    )
    state_cableLocked = 103  # Cable lock state [event] [Boolean] ['Admin', 'Partner', 'User']
    state_cableRating = 104  # Cable rating read [Amperes][event] [Integer] ['Admin', 'Partner', 'User']
    PilotHigh = 105  # Pilot signal high [Volt][debug] [Integer] ['Admin', 'Partner']
    PilotLow = 106  # Pilot signal low [Volt][debug] [Integer] ['Admin', 'Partner']
    state_backPlateID = 107  # Back Plate RFID of charger [boot] [String] ['Admin', 'Partner']
    state_userIDTokenReversed = 108  # User ID token string from RFID reading [event](NB! Must reverse these strings) [String] ['Admin', 'Partner']
    state_chargerOpMode = (
        109  # Charger operation mode according to charger mode table [event] [Integer] ['Admin', 'Partner', 'User']
    )
    state_outputPhase = 110  # Active output phase(s) to EV according to output phase type table. [event] [Integer] ['Admin', 'Partner', 'User']
    state_dynamicCircuitCurrentP1 = 111  # Dynamically set circuit maximum current for phase 1 [Amperes][event] [Double] ['Admin', 'Partner', 'User']
    state_dynamicCircuitCurrentP2 = 112  # Dynamically set circuit maximum current for phase 2 [Amperes][event] [Double] ['Admin', 'Partner', 'User']
    state_dynamicCircuitCurrentP3 = 113  # Dynamically set circuit maximum current for phase 3 [Amperes][event] [Double] ['Admin', 'Partner', 'User']
    state_outputCurrent = 114  # Available current signaled to car with pilot tone [Double] ['Admin', 'Partner', 'User']
    state_deratedCurrent = 115  # Available current after derating [A] [Double] ['Admin', 'Partner', 'User']
    state_deratingActive = 116  # Available current is limited by the charger due to high temperature [event] [Boolean] ['Admin', 'Partner', 'User']
    DebugString = 117  # Debug string [String] ['Admin', 'Partner']
    state_errorString = 118  # Descriptive error string [event] [String] ['Admin', 'Partner', 'User']
    state_errorCode = 119  # Error code according to error code table [event] [Integer] ['Admin', 'Partner', 'User']
    state_totalPower = 120  # Total power [kW][telemetry] [Double] ['Admin', 'Partner', 'User']
    state_sessionEnergy = 121  # Session accumulated energy [kWh][telemetry] [Double] ['Admin', 'Partner', 'User']
    state_energyPerHour = 122  # Accumulated energy per hour [kWh][event] [Double] ['Admin', 'Partner', 'User']
    LegacyEvStatus = 123  # 0 = not legacy ev, 1 = legacy ev detected, 2 = reviving ev [Integer] ['Admin', 'Partner']
    state_lifetimeEnergy = (
        124  # Accumulated energy in the lifetime of the charger [kWh] [Double] ['Admin', 'Partner', 'User']
    )
    state_lifetimeRelaySwitches = 125  # Total number of relay switches in the lifetime of the charger (irrespective of the number of phases used) [Integer] ['Admin', 'Partner']
    state_lifetimeHours = 126  # Total number of hours in operation [Integer] ['Admin', 'Partner']
    DynamicCurrentOfflineFallbackDEPRICATED = (
        127  # Maximum circuit current when offline [event] [Integer] ['Admin', 'Partner']
    )
    state_userIDToken = 128  # User ID token string from RFID reading [event] [String] ['Admin', 'Partner']
    state_ChargingSession = 129  # Charging sessions [json][event] [String] ['Admin', 'Partner']
    state_cellRSSI = 130  # Cellular signal strength [dBm][telemetry] [Integer] ['Admin', 'Partner', 'User']
    CellRAT = 131  # Cellular radio access technology according to RAT table [event] [Integer] ['Admin', 'Partner']
    state_wiFiRSSI = 132  # WiFi signal strength [dBm][telemetry] [Integer] ['Admin', 'Partner', 'User']
    CellAddress = 133  # IP address assigned by cellular network [debug] [String] ['Admin', 'Partner']
    WiFiAddress = 134  # IP address assigned by WiFi network [debug] [String] ['Admin', 'Partner']
    WiFiType = 135  # WiFi network type letters (G, N, AC, etc) [debug] [String] ['Admin', 'Partner']
    state_localRSSI = 136  # Local radio signal strength [dBm][telemetry] [Integer] ['Admin', 'Partner', 'User']
    MasterBackPlateID = 137  # Back Plate RFID of master [event] [String] ['Admin', 'Partner']
    LocalTxPower = 138  # Local radio transmission power [dBm][telemetry] [Integer] ['Admin', 'Partner']
    LocalState = 139  # Local radio state [event] [String] ['Admin', 'Partner']
    state_foundWiFi = 140  # List of found WiFi SSID and RSSI values [event] [String] ['Admin', 'Partner', 'User']
    state_chargerRAT = (
        141  # Radio access technology in use: 0 = cellular, 1 = wifi [Integer] ['Admin', 'Partner', 'User']
    )
    CellularInterfaceErrorCount = 142  # The number of times since boot the system has reported an error on this interface [poll] [Integer] ['Admin', 'Partner']
    CellularInterfaceResetCount = 143  # The number of times since boot the interface was reset due to high error count [poll] [Integer] ['Admin', 'Partner']
    WifiInterfaceErrorCount = 144  # The number of times since boot the system has reported an error on this interface [poll] [Integer] ['Admin', 'Partner']
    WifiInterfaceResetCount = 145  # The number of times since boot the interface was reset due to high error count [poll] [Integer] ['Admin', 'Partner']
    config_localNodeType = (
        146  # 0-Unconfigured, 1-Master, 2-Extender, 3-End device [Integer] ['Admin', 'Partner', 'User']
    )
    config_localRadioChannel = 147  # Channel nr 0 - 11 [Integer] ['Admin', 'Partner', 'User']
    config_localShortAddress = 148  # Address of charger on local radio network [Integer] ['Admin', 'Partner', 'User']
    config_localParentAddrOrNumOfNodes = (
        149  # If master-Number of slaves connected, If slave- Address parent [Integer] ['Admin', 'Partner', 'User']
    )
    state_tempMax = (
        150  # Maximum temperature for all sensors [Celsius][telemetry] [Integer] ['Admin', 'Partner', 'User']
    )
    state_tempAmbientPowerBoard = 151  # Temperature measured by ambient sensor on bottom of power board [Celsius][event] [Integer] ['Admin', 'Partner']
    state_tempInputT2 = 152  # Temperature at input T2 [Celsius][event] [Integer] ['Admin', 'Partner']
    state_tempInputT3 = 153  # Temperature at input T3 [Celsius][event] [Integer] ['Admin', 'Partner']
    state_tempInputT4 = 154  # Temperature at input T4 [Celsius][event] [Integer] ['Admin', 'Partner']
    state_tempInputT5 = 155  # Temperature at input T5 [Celsius][event] [Integer] ['Admin', 'Partner']
    state_tempOutputN = (
        160  # Temperature at type 2 connector plug for N [Celsius][event] [Integer] ['Admin', 'Partner']
    )
    state_tempOutputL1 = (
        161  # Temperature at type 2 connector plug for L1 [Celsius][event] [Integer] ['Admin', 'Partner']
    )
    state_tempOutputL2 = (
        162  # Temperature at type 2 connector plug for L2 [Celsius][event] [Integer] ['Admin', 'Partner']
    )
    state_tempOutputL3 = (
        163  # Temperature at type 2 connector plug for L3 [Celsius][event] [Integer] ['Admin', 'Partner']
    )
    state_tempRelayN = 164  # Temperature under N relay on ONE [Celsius] [Integer] ['Admin', 'Partner']
    state_tempRelayL = 165  # Temperature under L relay on ONE [Celsius] [Integer] ['Admin', 'Partner']
    state_tempAmbientPowerBoardTop = (
        166  # Temperature measured by ambient sensor on top of power board [Celsius] [Integer] ['Admin', 'Partner']
    )
    state_tempAmbient = 170  # Ambient temperature on COM board [Celsius][event] [Integer] ['Admin', 'Partner']
    state_lightAmbient = 171  # Ambient light from front side [Percent][debug] [Integer] ['Admin', 'Partner']
    state_intRelHumidity = 172  # Internal relative humidity [Percent][event] [Integer] ['Admin', 'Partner']
    state_backPlateLocked = 173  # Back plate confirmed locked [event] [Boolean] ['Admin', 'Partner']
    state_currentMotor = 174  # Motor current draw [debug] [Double] ['Admin', 'Partner']
    state_backPlateHallSensor = 175  # Raw sensor value [mV] [Integer] ['Admin', 'Partner']
    state_inCurrentT2 = (
        182  # Calculated current RMS for input T2 [Amperes][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inCurrentT3 = 183  # Current RMS for input T3 [Amperes][telemetry] [Double] ['Admin', 'Partner', 'User']
    state_inCurrentT4 = 184  # Current RMS for input T4 [Amperes][telemetry] [Double] ['Admin', 'Partner', 'User']
    state_inCurrentT5 = 185  # Current RMS for input T5 [Amperes][telemetry] [Double] ['Admin', 'Partner', 'User']
    state_inVoltageT1T2 = (
        190  # Input voltage RMS between T1 and T2 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT1T3 = (
        191  # Input voltage RMS between T1 and T3 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT1T4 = (
        192  # Input voltage RMS between T1 and T4 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT1T5 = (
        193  # Input voltage RMS between T1 and T5 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT2T3 = (
        194  # Input voltage RMS between T2 and T3 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT2T4 = (
        195  # Input voltage RMS between T2 and T4 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT2T5 = (
        196  # Input voltage RMS between T2 and T5 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT3T4 = (
        197  # Input voltage RMS between T3 and T4 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT3T5 = (
        198  # Input voltage RMS between T3 and T5 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_inVoltageT4T5 = (
        199  # Input voltage RMS between T4 and T5 [Volt][telemetry] [Double] ['Admin', 'Partner', 'User']
    )
    state_nominalVoltage = 200  # Nominal voltage setting for Easee One [Integer] ['Admin', 'Partner', 'User']
    state_outVoltPin1to2 = (
        202  # Output voltage RMS between type 2 pin 1 and 2 [Volt][telemetry] [Double] ['Admin', 'Partner']
    )
    state_outVoltPin1to3 = (
        203  # Output voltage RMS between type 2 pin 1 and 3 [Volt][telemetry] [Double] ['Admin', 'Partner']
    )
    state_outVoltPin1to4 = (
        204  # Output voltage RMS between type 2 pin 1 and 4 [Volt][telemetry] [Double] ['Admin', 'Partner']
    )
    state_outVoltPin1to5 = (
        205  # Output voltage RMS between type 2 pin 1 and 5 [Volt][telemetry] [Double] ['Admin', 'Partner']
    )
    state_outVoltPin2_3 = 206  # Output voltage RMS between type 2 pin 2 and 3 [Volt] [Double] ['Admin', 'Partner']
    state_voltLevel33 = 210  # 3.3 Volt Level [Volt][telemetry] [Integer] ['Admin', 'Partner']
    state_voltLevel5 = 211  # 5 Volt Level [Volt][telemetry] [Integer] ['Admin', 'Partner']
    state_voltLevel12 = 212  # 12 Volt Level [Volt][telemetry] [Integer] ['Admin', 'Partner']
    state_fatalErrorCode = (
        219  # Fatal error code according to error code table [event] [Integer] ['Admin', 'Partner', 'User']
    )
    state_LTERSRP = 220  # Reference Signal Received Power (LTE) [-144 .. -44 dBm] [Integer] ['Admin', 'Partner']
    state_LTESINR = 221  # Signal to Interference plus Noise Ratio (LTE) [-20 .. +30 dB] [Integer] ['Admin', 'Partner']
    state_LTERSRQ = 222  # Reference Signal Received Quality (LTE) [-19 .. -3 dB] [Integer] ['Admin', 'Partner']
    state_chargingSessionStart = 223  # Charging session started [event] [String] ['Admin', 'Partner']
    state_eqAvailableCurrentP1 = (
        230  # Available current for charging on P1 according to Equalizer [Double] ['Admin', 'Partner', 'User']
    )
    state_eqAvailableCurrentP2 = (
        231  # Available current for charging on P2 according to Equalizer [Double] ['Admin', 'Partner', 'User']
    )
    state_eqAvailableCurrentP3 = (
        232  # Available current for charging on P3 according to Equalizer [Double] ['Admin', 'Partner', 'User']
    )
    state_diagnosticsString = 240  # Various diagnostic information for testing. Must be enabled by setting DiagnosticsMode == 256 [String] ['Admin', 'Partner']
    state_connectedToCloud = 250  # Device is connected to AWS [Boolean] ['Admin', 'Partner', 'User']
    state_cloudDisconnectReason = 251  # AWS DisconnectReason [String] ['Admin', 'Partner', 'User']


# Data IDs for Equalizer from signalr stream
class EqualizerStreamData(Enum):
    state_selfTestResult = 1  # PASSED or error codes [String] ['Admin', 'Partner']
    state_selfTestDetails = 2  # JSON with details from self-test [String] ['Admin', 'Partner']
    state_easeeLinkCommandResponse = (
        13  # Response on a EaseeLink command sent to another devic [Integer] ['Admin', 'Partner']
    )
    state_easeeLinkDataReceived = 14  # Data received on EaseeLink from another device [String] ['Admin', 'Partner']
    state_siteIDNumeric = 19  # Site ID numeric value [event] [Integer] ['Admin', 'Partner']
    config_siteStructure = 20  # Site Structure [boot] [String] ['Admin', 'Partner', 'User']
    state_softwareRelease = 21  # Embedded software package release id [boot] [Integer] ['Admin', 'Partner', 'User']
    state_deviceMode = 23  # Current device mode [Integer] ['Admin', 'Partner', 'User']
    config_meterType = 25  # Meter type [String] ['Admin', 'Partner', 'User']
    config_meterID = 26  # Meter identification [String] ['Admin', 'Partner', 'User']
    state_OBISListIdentifier = 27  # OBIS List version identifier [String] ['Admin', 'Partner']
    config_gridType = 29  # 0=Unknown, 1=TN, 2=IT, [Integer] ['Admin', 'Partner', 'User']
    config_numPhases = 30  # [Integer] ['Admin', 'Partner', 'User']
    state_currentL1 = 31  # Current in Amps [Double] ['Admin', 'Partner', 'User']
    state_currentL2 = 32  # Current in Amps [Double] ['Admin', 'Partner', 'User']
    state_currentL3 = 33  # Current in Amps [Double] ['Admin', 'Partner', 'User']
    state_voltageNL1 = 34  # Voltage in Volts [Double] ['Admin', 'Partner', 'User']
    state_voltageNL2 = 35  # Voltage in Volts [Double] ['Admin', 'Partner', 'User']
    state_voltageNL3 = 36  # Voltage in Volts [Double] ['Admin', 'Partner', 'User']
    state_voltageL1L2 = 37  # Voltage in Volts [Double] ['Admin', 'Partner', 'User']
    state_voltageL1L3 = 38  # Voltage in Volts [Double] ['Admin', 'Partner', 'User']
    state_voltageL2L3 = 39  # Voltage in Volts [Double] ['Admin', 'Partner', 'User']
    state_activePowerImport = 40  # Active Power Import in kW [Double] ['Admin', 'Partner', 'User']
    state_activePowerExport = 41  # Active Power Export in kW [Double] ['Admin', 'Partner', 'User']
    state_reactivePowerImport = 42  # Reactive Power Import in kVAR [Double] ['Admin', 'Partner', 'User']
    state_reactivePowerExport = 43  # Reactive Power Export in kVAR [Double] ['Admin', 'Partner', 'User']
    state_maxPowerImport = 44  # Maximum power import[event] [Double] ['Admin', 'Partner', 'User']
    state_cumulativeActivePowerImport = (
        45  # Cumulative Active Power Import in kWh [Double] ['Admin', 'Partner', 'User']
    )
    state_cumulativeActivePowerExport = (
        46  # Cumulative Active Power Export in kWh [Double] ['Admin', 'Partner', 'User']
    )
    state_cumulativeReactivePowerImport = (
        47  # Cumulative Reactive Power Import in kVARh [Double] ['Admin', 'Partner', 'User']
    )
    state_cumulativeReactivePowerExport = (
        48  # Cumulative Reactive Power Export in kVARh [Double] ['Admin', 'Partner', 'User']
    )
    state_clockAndDateMeter = 49  # Clock and Date from Meter [String] ['Admin', 'Partner', 'User']
    state_rcpi = 50  # Received Channel Power Indicator (dBm) [Double] ['Admin', 'Partner', 'User']
    config_ssid = 51  # WIFI SSID  [String] ['Admin', 'Partner', 'User']
    config_masterBackPlateID = 55  # Back Plate RFID of master charger [event] [String] ['Admin', 'Partner', 'User']
    config_equalizerID = 56  # Back Plate RFID of equalizer [boot] [String] ['Admin', 'Partner', 'User']
    config_childReport = 57  # Child configuration in Equalizer [String] ['Admin', 'Partner', 'User']
    state_connectivityReport = 58  # Child connectivity [String] ['Admin', 'Partner']
    state_exceptionData = 60  # Exception Debug Information [boot] [String] ['Admin', 'Partner']
    state_bootReason = 61  # (Re)boot reason [boot] [String] ['Admin', 'Partner']
    state_highCurrentTransitions = (
        64  # Number of transitions to high current mode (Debug) [Integer] ['Admin', 'Partner']
    )
    state_vCap = 65  # Capacitor Voltage in Volts [Double] ['Admin', 'Partner']
    state_vBusMin = 66  # Minimum Bus Voltage in Volts [Double] ['Admin', 'Partner']
    state_vbusMax = 67  # Maximum Bus Voltage in Volts [Double] ['Admin', 'Partner']
    state_internalTemperature = 68  # Internal temperature in Celsius [Double] ['Admin', 'Partner']
    state_meterDataSnapshot = 69  # Meter data snapshot [String] ['Admin', 'Partner']
    state_localRSSI = 70  # Local radio signal strength [dBm][telemetry] [Integer] ['Admin', 'Partner', 'User']
    state_localTxPower = 71  # Local radio transmission power [dBm][telemetry] [Integer] ['Admin', 'Partner']
    state_localRadioChannel = 72  # Local radio channel nr 0 - 11 [telemetry] [Integer] ['Admin', 'Partner']
    state_localShortAddress = (
        73  # Address of equalizer on local radio network [telemetry] [Integer] ['Admin', 'Partner']
    )
    state_localNodeType = 74  # 0-Unconfigured, 1 - Coordinator, 2 - Range extender, 3 - End device, 4- Sleepy end device [telemetry] [Integer] ['Admin', 'Partner']
    state_localParentAddress = 75  # Address of parent on local radio network. If 0 - master, else extender [telemetry] [Integer] ['Admin', 'Partner']
    config_circuitPhaseMapping = 80  # Mapping between EQ phases and charger phases [String] ['Admin', 'Partner']
    state_phaseMappingReport = 81  # Charger vs Meter phase correlation report [String] ['Admin', 'Partner']
    state_phaseLearningStatus = 82  # Status of the phaselearning/loadbalancing [String] ['Admin', 'Partner']
    config_modbusConfiguration = 85  # Complete Modbus Configuration [String] ['Admin', 'Partner', 'User']
    state_loadbalanceThrottle = 86  # Throttle level [percent] [Integer] ['Admin', 'Partner']
    state_availableCurrentL1 = 87  # Available Current for Balancing in Amps [Double] ['Admin', 'Partner', 'User']
    state_availableCurrentL2 = 88  # Available Current for Balancing in Amps [Double] ['Admin', 'Partner', 'User']
    state_availableCurrentL3 = 89  # Available Current for Balancing in Amps [Double] ['Admin', 'Partner', 'User']
    state_meterErrors = 90  # Meter Errors [Integer] ['Admin', 'Partner']
    state_APMacAddress = 91  # Mac Address of the Wifi access point [String] ['Admin', 'Partner']
    state_wifiReconnects = 92  # Number of sucessful reconnects to AP [Integer] ['Admin', 'Partner']
    state_ledMode = 100  # Current LED pattern [Integer] ['Admin', 'Partner', 'User']
    state_equalizedChargeCurrentL1 = (
        105  # Charge current controlled by Equalizer in Amps [Double] ['Admin', 'Partner', 'User']
    )
    state_equalizedChargeCurrentL2 = (
        106  # Charge current controlled by Equalizer in Amps [Double] ['Admin', 'Partner', 'User']
    )
    state_equalizedChargeCurrentL3 = (
        107  # Charge current controlled by Equalizer in Amps [Double] ['Admin', 'Partner', 'User']
    )
    config_currentTransformerConfig = 110  # Current Transformer Configuration [String] ['Admin', 'Partner', 'User']
    state_meterEncryptionStatus = 111  # Meter Encryption Status [String] ['Admin', 'Partner', 'User']
    state_surplusCharging = 115  # Surplus charging configuration [String] ['Admin', 'Partner', 'User']
    state_connectedAmps = 120  # Equalizer AMPs report ['Admin', 'Partner', 'User']
    state_connectedToCloud = 250  # Device is connected to AWS [Boolean] ['Admin', 'Partner', 'User']
    state_cloudDisconnectReason = 251  # AWS DisconnectReason [String] ['Admin', 'Partner', 'User']


# Data IDs for Equalizer from signalr stream
class DatatypesStreamData(Enum):
    Binary = 1
    Boolean = 2
    Double = 3
    Integer = 4
    Position = 5
    String = 6
    Statistics = 7



from datetime import datetime
import logging
from typing import Any, Dict, List

from .charger import Charger, ChargerConfig, ChargerState
from .exceptions import ForbiddenServiceException, ServerFailureException
from .utils import BaseDict

_LOGGER = logging.getLogger(__name__)


class EqualizerState(BaseDict):
    def __init__(self, state: Dict[str, Any]):
        data = {**state}
        super().__init__(data)


class EqualizerConfig(BaseDict):
    def __init__(self, config: Dict[str, Any]):
        data = {**config}
        super().__init__(data)


class Equalizer(BaseDict):
    def __init__(self, data: Dict[str, Any], site: Any, easee: Any):
        super().__init__(data)
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.site = site
        self.easee = easee

    async def get_observations(self, *args):
        """Gets observation IDs"""
        observation_ids = ",".join(str(s) for s in args)
        try:
            return await (await self.easee.get(f"/state/{self.id}/observations?ids={observation_ids}")).json()
        except (ServerFailureException):
            return None

    async def get_state(self):
        """Get Equalizer state"""
        try:
            state = await (await self.easee.get(f"/api/equalizers/{self.id}/state")).json()
            return EqualizerState(state)
        except (ServerFailureException):
            return None

    async def get_config(self):
        """Get Equalizer config"""
        try:
            config = await (await self.easee.get(f"/api/equalizers/{self.id}/config")).json()
            return EqualizerConfig(config)
        except (ServerFailureException):
            return None

    async def empty_state(self, raw=False):
        """Create an empty state data structyre"""
        state = {}
        return EqualizerState(state)

    async def empty_config(self, raw=False):
        """Crate an empty config data structure"""
        config = {}
        return EqualizerConfig(config)

    async def get_latest_firmware(self):
        """Get the latest released firmeware version"""
        try:
            return await (await self.easee.get(f"/firmware/{self.id}/latest")).json()
        except (ServerFailureException):
            return None


class Circuit(BaseDict):
    """Represents a Circuit"""

    def __init__(self, data: Dict[str, Any], site: Any, easee: Any):
        super().__init__(data)
        self.id: int = data["id"]
        self.site = site
        self.easee = easee

    async def set_dynamic_current(
        self, currentP1: int, currentP2: int = None, currentP3: int = None, timeToLive: int = 0
    ):
        """Set dynamic current on circuit level. timeToLive specifies, in minutes, for how long the new dynamic current is valid. timeToLive = 0 means that the new dynamic current is valid until changed the next time. The dynamic current is always reset to default when the charger is restarted"""
        json = {
            "phase1": currentP1,
            "phase2": currentP2 if currentP2 is not None else currentP1,
            "phase3": currentP3 if currentP3 is not None else currentP1,
            "timeToLive": timeToLive,
        }
        try:
            return await self.easee.post(f"/api/sites/{self.site.id}/circuits/{self.id}/dynamicCurrent", json=json)
        except (ServerFailureException):
            return None

    async def set_max_current(self, currentP1: int, currentP2: int = None, currentP3: int = None):
        """Set circuit max current"""
        json = {
            "maxCircuitCurrentP1": currentP1,
            "maxCircuitCurrentP2": currentP2 if currentP2 is not None else currentP1,
            "maxCircuitCurrentP3": currentP3 if currentP3 is not None else currentP1,
        }
        try:
            return await self.easee.post(f"/api/sites/{self.site.id}/circuits/{self.id}/settings", json=json)
        except (ServerFailureException):
            return None

    async def set_max_offline_current(self, currentP1: int, currentP2: int = None, currentP3: int = None):
        """Set circuit max offline current, fallback value for limit if charger is offline"""
        json = {
            "offlineMaxCircuitCurrentP1": currentP1,
            "offlineMaxCircuitCurrentP2": currentP2 if currentP2 is not None else currentP1,
            "offlineMaxCircuitCurrentP3": currentP3 if currentP3 is not None else currentP1,
        }
        try:
            return await self.easee.post(f"/api/sites/{self.site.id}/circuits/{self.id}/settings", json=json)
        except (ServerFailureException):
            return None

    async def set_rated_current(self, ratedCurrentFuseValue: int):
        """Set circuit rated current - requires elevated access (installers only)"""
        json = {"ratedCurrentFuseValue": ratedCurrentFuseValue}
        try:
            return await self.easee.post(f"/api/sites/{self.site.id}/circuits/{self.id}/rated_current", json=json)
        except (ServerFailureException):
            return None

    def get_chargers(self) -> List[Charger]:
        return [Charger(c, self.easee, self.site, self) for c in self["chargers"]]


class SiteState(BaseDict):
    """Represents the site state as received through /sites/<id>/state"""

    def __init__(self, data: Dict[str, Any]):
        super().__init__(data)

    def get_charger_config(self, charger_id: str, raw=False) -> ChargerConfig:
        """get config for charger from the instance data"""
        for circuit in self["circuitStates"]:
            for charger_data in circuit["chargerStates"]:
                if charger_data["chargerID"] == charger_id:
                    return ChargerConfig(charger_data["chargerConfig"], raw)

        return None

    def get_charger_state(self, charger_id: str, raw=False) -> ChargerState:
        """get state for charger from the instance data"""
        for circuit in self["circuitStates"]:
            for charger_data in circuit["chargerStates"]:
                if charger_data["chargerID"] == charger_id:
                    return ChargerState(charger_data["chargerState"], raw)

        return None


class Site(BaseDict):
    """Represents a Site"""

    def __init__(self, data: Dict[str, Any], easee: Any):
        super().__init__(data)
        self.id: int = data["id"]
        self.name: str = data["name"]
        self.easee = easee

    def get_circuits(self) -> List[Circuit]:
        """Get circuits for the site"""
        return [Circuit(c, self, self.easee) for c in self["circuits"]]

    def get_equalizers(self) -> List[Equalizer]:
        """Get equalizers for the site"""
        return [Equalizer(e, self, self.easee) for e in self["equalizers"]]

    async def set_name(self, name: str):
        """Set name for the site"""
        json = {**self.get_data(), "name": name}
        return await self.easee.put(f"/api/sites/{self.id}", json=json)

    async def set_currency(self, currency: str):
        """Set currency for the site"""
        json = {**self.get_data(), "currencyId": currency}
        val = await self.easee.put(f"/api/sites/{self.id}", json=json)
        self["currencyId"] = currency
        return val

    async def set_price(
        self,
        costPerKWh: float,
        vat: float = None,
        currency: str = None,
        costPerKwhExcludeVat: float = None,
    ):
        """Set price per kWh for the site"""

        json = {"costPerKWh": costPerKWh}

        if vat is None:
            vat = self.get("vat")

        if currency is None:
            currency = self.get("currencyId")

        if costPerKwhExcludeVat is None:
            costPerKwhExcludeVat = costPerKWh / (100.0 + vat) * 100.0

        json = {
            "currencyId": currency,
            "costPerKWh": costPerKWh,
            "vat": vat,
            "costPerKwhExcludeVat": costPerKwhExcludeVat,
        }

        try:
            val = await self.easee.post(f"/api/sites/{self.id}/price", json=json)
            self["vat"] = vat
            self["currencyId"] = currency
            self["costPerKWh"] = costPerKWh
            self["costPerKwhExcludeVat"] = costPerKwhExcludeVat
            return val
        except (ServerFailureException):
            return None

    async def get_cost_between_dates(self, from_date: datetime, to_date: datetime):
        """Get the charging cost between from_datetime and to_datetime"""

        try:
            costs = await (
                await self.easee.get(f"/api/sites/{self.id}/breakdown/{from_date.isoformat()}/{to_date.isoformat()}")
            ).json()
            return costs
        except (ServerFailureException, ForbiddenServiceException):
            return None



from collections.abc import Mapping
from datetime import datetime, timezone
import re

from .const import ChargerStreamData, EqualizerStreamData

regex = r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"
match_iso8601 = re.compile(regex).match


def lookup_charger_stream_id(id):
    try:
        name = ChargerStreamData(id).name
    except ValueError:
        return None
    return name


def lookup_equalizer_stream_id(id):
    try:
        name = EqualizerStreamData(id).name
    except ValueError:
        return None
    return name


def convert_stream_data(data_type, value):
    if data_type == 2:
        if value.lower() in ["1", "true", "on", "yes"]:
            return True
        else:
            return False
    elif data_type == 3:
        return float(value)
    elif data_type == 4:
        return int(value)

    return value


def validate_iso8601(str_val):
    try:
        if match_iso8601(str_val) is not None:
            return True
    except Exception:
        pass
    return False


class BaseDict(Mapping):
    def __init__(self, entries):
        self._storage = entries

    def __getitem__(self, key):
        if type(self._storage[key]) == str and validate_iso8601(self._storage[key]):
            try:
                return datetime.fromisoformat(self._storage[key]).replace(tzinfo=timezone.utc)
            except ValueError:
                try:
                    return datetime.strptime(self._storage[key], "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
                except ValueError:
                    (dt, msecs) = self._storage[key].split(".")
                    return datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
        return self._storage[key]

    def __setitem__(self, key, value):
        self._storage[key] = value

    def __iter__(self):
        return iter(self._storage)  # ``ghost`` is invisible

    def __len__(self):
        return len(self._storage)

    def get_data(self):
        return self._storage
